import numpy as np
from . import Geometry

class Rectangle(Geometry.Geometry):
    def __init__(self, **kwargs):
        Geometry.Geometry.__init__(self, **kwargs)
        self.shape = "rect"
        self.x = kwargs['x']
        self.y = kwargs['y']
        self.rx = kwargs.get('rx', "0")
        self.ry = kwargs.get('ry', "0")
        self.width = kwargs['width']
        self.height = kwargs['height']
        self.non = kwargs['non']
        self.isClosed = True

        w = self.width
        h = self.height
        non = self.non
        edge1x = np.arange(-w/2, w/2, w/non)
        edge1y = np.arange(len(edge1x))*0 - h/2
        edge2y = np.arange(-h/2, h/2, h/(non*h/w))
        edge2x = np.arange(len(edge2y))*0 + w/2
        
        edge3x = np.arange(w/2, -w/2, -w/non)
        edge3y = np.arange(len(edge3x))*0 + h/2
        edge4y = np.arange(h/2, -h/2, -h/(non*h/w))
        edge4x = np.arange(len(edge4y))*0 - w/2
        
        x = np.hstack([edge1x, edge2x, edge3x, edge4x])
        y = np.hstack([edge1y, edge2y, edge3y, edge4y])

        # list of points
        v0 = np.vstack([x, y])*float(self.scale)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.x], [self.y]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array([self.x, self.y])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        segments.append([len(boundaryNodes)-1, 0])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        plt.plot(v[[0,i+1],0], v[[0,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))








import numpy as np

from scipy.spatial import Delaunay as delaunay
from shapely.geometry.polygon import Polygon as ShapelyPolygon
from shapely.geometry import Point as ShapelyPoint

class Mesh:
    def __init__(self, x, y, conmat):
        self.x = x
        self.y = y
        self.conmat = conmat

    def getNeiElements(self):
        nn = len(self.x)
        ne = len(self.conmat[:,0])
        neiElements = []
        for i in range(nn):
            ele = []
            for j in range(ne):
                if self.conmat[j,0] == i or self.conmat[j,1] == i or self.conmat[j,2] == i:
                    ele.append(j)
            neiElements.append(ele)
        self.neiElements = neiElements

    def getNeiNodes(self):
        nn = len(self.x)
        neiNodes = []
        for i in range(nn):
            nod = []
            for j in self.neiElements[i]:
                for k in self.conmat[j]:
                    if k != i: 
                        nod.append(k)
            nod = np.unique(np.array(nod))
            nod = nod.tolist()
            neiNodes.append(nod)
        self.neiNodes = neiNodes
        
def prepareGeometryForMeshing(shapes):
    # Finding all child shapes
    for i in range(len(shapes)): # for all shapes,
        if shapes[i].isClosed: # if the current shape is closed, it may contain a child.
            for j in range(len(shapes)): # for all potential child shapes,
                if i != j: # skip current shape as potential child.
                    polygon = ShapelyPolygon(shapes[i].boundaryNodes) # consider current shape as polygon,
                    if shapes[j].shape == "point":
                        point = ShapelyPoint(shapes[j].boundaryNodes) # and take one point form potential child shape.
                    else:
                        point = ShapelyPoint(shapes[j].boundaryNodes[0]) # and take one point form potential child shape.
                    if polygon.contains(point): # if one point of potential child shape is inside current shape,
                        shapes[i].child.append(j) # potential child shape is a child of current shape. Note: all the shapes must be non-contacting.

    # Finding all parent shapes
    for i in range(len(shapes)): # for all shapes,
        for j in range(len(shapes)): # for all potential parent shapes,
            if i != j: # skip current shape as potential parent.
                if shapes[j].isClosed: # if the potential shape is closed, it may be a parent.
                    polygon = ShapelyPolygon(shapes[j].boundaryNodes) # consider potential parent shape as polygon,
                    if shapes[i].shape == "point":
                        point = ShapelyPoint(shapes[i].boundaryNodes) # and take one point from current shape.
                    else:
                        point = ShapelyPoint(shapes[i].boundaryNodes[0]) # and take one point from current shape.
                    if polygon.contains(point): # if one point of current shape is inside the potential parent shape,
                        shapes[i].parent.append(j) # potential parent shape is the parent of current shape. Note: all the shapes must be non-contacting.

    for i in range(len(shapes)):
        shapes[i].parent = np.array(shapes[i].parent)
        shapes[i].child = np.array(shapes[i].child)

    # Finding immediate child shapes
    listOfShapesToRemove = []
    for i in range(len(shapes)):
        shapesToRemove = []
        for j in range(len(shapes[i].child)):
            for k in range(len(shapes[shapes[i].child[j]].child)):
                shapesToRemove.append(shapes[shapes[i].child[j]].child[k])
        listOfShapesToRemove.append(shapesToRemove)

    for i in range(len(shapes)):
        for j in range(len(listOfShapesToRemove[i])):
            shapes[i].child = np.delete(shapes[i].child, np.where(shapes[i].child == listOfShapesToRemove[i][j]))#, axis=0)

    # Finding immediate parent shapes
    listOfShapesToRemove = []
    for i in range(len(shapes)):
        shapesToRemove = []
        for j in range(len(shapes[i].parent)):
            for k in range(len(shapes[shapes[i].parent[j]].parent)):
                shapesToRemove.append(shapes[shapes[i].parent[j]].parent[k])
        listOfShapesToRemove.append(shapesToRemove)
        
    for i in range(len(shapes)):
        for j in range(len(listOfShapesToRemove[i])):
            shapes[i].parent = np.delete(shapes[i].parent, np.where(shapes[i].parent ==listOfShapesToRemove[i][j]))#, axis=0)

    # Labelling if a shape is child or parent
    for i in range(len(shapes)):
        if len(shapes[i].child) == 0:
            shapes[i].isParent = False
        else:
            shapes[i].isParent = True

    for i in range(len(shapes)):
        if len(shapes[i].parent) == 0:
            shapes[i].isChild = False
        else:
            shapes[i].isChild = True

def generateNodes(shapes, crse, meshingMethod):
    # Generate interior free nodes
    for i in range(len(shapes)):
        if shapes[i].isClosed:
            if meshingMethod == "ruppert":
                meshShapeRuppert(i, shapes, crse)
            elif meshingMethod == "triangle":
                meshShapeTriangle(i, shapes, crse)
        else:
            shapes[i].freeNodes = np.array([])

def meshShapeTriangle(ind, shapes):
    # Parent points
    parentPoints = shapes[ind].boundaryNodes
    # Child points
    childPoints = []
    for i in range(len(shapes[ind].child)):
        childPoints.append(shapes[shapes[ind].child[i]].boundaryNodes.tolist())
    
    vertices = parentPoints.append(childPoints)

    # Parent segments
    segments = []
    for i in range(len(parentPoints)-1):
        segments.append([i, i+1])
    segments.append([len(parentPoints)-1, 0])
    # Child segments
    segInd = len(parentPoints)
    for i in range(len(shapes[ind].child)):
        for j in range(len(childPoints[i])-1):
            if shapes[shapes[ind].child[i]].shape != "point":
                segments.append([segInd+j, segInd+j+1])
        if shapes[shapes[ind].child[i]].isClosed:
            segments.append([segInd+j+1, segInd])
        if shapes[shapes[ind].child[i]].shape == 'point':
            segInd = segInd + 1
        else:
            segInd = segInd + len(childPoints[i])

    regions = []


def meshShapeRuppert(ind, shapes, crse): # Consider skipping sengment initialisation in case of Point
    # Parent points
    parentPoints = shapes[ind].boundaryNodes
    # Child points
    childPoints = []
    for i in range(len(shapes[ind].child)):
        childPoints.append(shapes[shapes[ind].child[i]].boundaryNodes.tolist())
    
    # Parent polygon
    parentPolygon = ShapelyPolygon(parentPoints)
    # Child polygons
    childPolygons = []
    for i in range(len(shapes[ind].child)):
        if shapes[shapes[ind].child[i]].isClosed:
            childPolygons.append(ShapelyPolygon(childPoints[i]))

    # Parent segments
    segments = []
    for i in range(len(parentPoints)-1):
        segments.append([i, i+1])
    segments.append([len(parentPoints)-1, 0])
    # Child segments
    segInd = len(parentPoints)
    for i in range(len(shapes[ind].child)):
        for j in range(len(childPoints[i])-1):
            if shapes[shapes[ind].child[i]].shape != "point":
                segments.append([segInd+j, segInd+j+1])
        if shapes[shapes[ind].child[i]].isClosed:
            segments.append([segInd+j+1, segInd])
        if shapes[shapes[ind].child[i]].shape == 'point':
            segInd = segInd + 1
        else:
            segInd = segInd + len(childPoints[i])

    # Points
    points = parentPoints

    for i in range(len(childPoints)):
        points = np.vstack((points,childPoints[i]))

    dt = delaunay(points)
    cm = dt.simplices  # Connectivity matrix
    x = points[:, 0]
    y = points[:, 1]
    ix = x.copy()
    iy = y.copy()

    elementsToRemove = []
    for i in range(len(cm)):
        cot = ShapelyPoint((x[cm[i][0]]+x[cm[i][1]]+x[cm[i][2]])/3, (y[cm[i][0]]+y[cm[i][1]]+y[cm[i][2]])/3)
        if not parentPolygon.contains(cot):
            elementsToRemove.append(i)
        for childPolygon in childPolygons:
            if childPolygon.contains(cot):
                elementsToRemove.append(i)
    cm = np.delete(cm, elementsToRemove, 0)
    st = getSharpTriangles(x, y, cm, crse)

    while len(st) != 0:
        for element in st:
            ex = x[cm[element]]
            ey = y[cm[element]]
            cc = circumCenter(ex, ey)

            isNotEcroached = segmentsEncroached(points, segments, cc)
            cirCen = ShapelyPoint(cc[0], cc[1])
            if parentPolygon.contains(cirCen) and isNotEcroached:
                pointOutsideChild = True
                for i in range(len(childPolygons)):
                    if childPolygons[i].contains(cirCen):
                        pointOutsideChild = False
                        break
                if pointOutsideChild:
                    points = np.vstack([points, cc])
                    dt = delaunay(points)
                    cm = dt.simplices # Connectivity matrix
                    x = points[:,0]
                    y = points[:,1]

                    elementsToRemove = []
                    for i in range(len(cm)):
                        cot = ShapelyPoint((x[cm[i][0]]+x[cm[i][1]]+x[cm[i][2]])/3, (y[cm[i][0]]+y[cm[i][1]]+y[cm[i][2]])/3)
                        if not parentPolygon.contains(cot):
                            elementsToRemove.append(i)
                        for childPolygon in childPolygons:
                            if childPolygon.contains(cot):
                                elementsToRemove.append(i)
                    cm = np.delete(cm, elementsToRemove, 0)
                    break
        stNew = getSharpTriangles(x, y, cm, crse)
        if stNew == st:
            break
        else:
            st = stNew

    fx = x[len(ix):len(x)-1]
    fy = y[len(iy):len(y)-1]
    freeNodes = []
    for i in range(len(fx)):
        freeNodes.append([fx[i],fy[i]])
    freeNodes = np.array(freeNodes)
    shapes[ind].freeNodes = freeNodes

def circumCenter(ex, ey):
    Ax, Ay, Bx, By, Cx, Cy = ex[0], ey[0], ex[1], ey[1], ex[2], ey[2]
    D = 2*(Ax*(By-Cy)+Bx*(Cy-Ay)+Cx*(Ay-By))
    cx = ((Ax*Ax+Ay*Ay)*(By-Cy)+(Bx*Bx+By*By)*(Cy-Ay)+(Cx*Cx+Cy*Cy)*(Ay-By))/D
    cy = ((Ax*Ax+Ay*Ay)*(Cx-Bx)+(Bx*Bx+By*By)*(Ax-Cx)+(Cx*Cx+Cy*Cy)*(Bx-Ax))/D
    return cx, cy

def circumRadiusToMinSideRatio(ex, ey):
    a, b, c = np.sqrt((ex[1]-ex[0])**2+(ey[1]-ey[0])**2), np.sqrt((ex[2]-ex[1])**2+(ey[2]-ey[1])**2), np.sqrt((ex[2]-ex[0])**2+(ey[2]-ey[0])**2)
    minSide = min([a, b, c])
    cr = (a*b*c)/(np.sqrt((a + b + c)*(b + c - a)*(c + a - b)*(a + b - c))) # Circumradius
    return cr/minSide

def getSharpTriangles(x, y, cm, B):
    st = []
    for i in range(len(cm)):
        element = cm[i]
        ex = x[element]
        ey = y[element]
        circumRatio = circumRadiusToMinSideRatio(ex, ey)
        if circumRatio > B:
            st.append(i)
    return st

def encroached(p1, p2, p):
    m = (np.array(p1)+np.array(p2))/2
    r = np.sqrt((p1[0]-m[0])**2+(p1[1]-m[1])**2)
    d = np.sqrt((p[0]-m[0])**2+(p[1]-m[1])**2)
    if d > r:
        return False
    else:
        return True

def segmentsEncroached(points, segments, cc):
    isNotEncroached = True
    for i in range(len(segments)):
        segmentCentre = [(points[segments[i][0]][0]+points[segments[i][1]][0])/2 ,(points[segments[i][0]][1]+points[segments[i][1]][1])/2]
        segmentRadius = np.sqrt((points[segments[i][0]][0]-points[segments[i][1]][0])**2+(points[segments[i][0]][1]-points[segments[i][1]][1])**2)/2
        distanceFromCentre = np.sqrt((cc[0]-segmentCentre[0])**2+(cc[1]-segmentCentre[1])**2)
        if distanceFromCentre < segmentRadius:
            isNotEncroached = False
            break
    return isNotEncroached

def createMesh(shapes):

    x = np.array([])
    y = np.array([])
    nn = -1

    for shape in shapes:
        if shape.shape == "point":
            shape.boundaryList = np.array([nn+1])
            nn = nn+1
            shape.freeList = np.array([])
            x =  np.concatenate((x, [shape.boundaryNodes[0]]))
            y =  np.concatenate((y, [shape.boundaryNodes[1]]))

        elif shape.shape == "line" or shape.shape == "polyline":
            shape.boundaryList = np.arange(nn+1, nn+len(shape.boundaryNodes[:, 0])+1, 1)
            nn = shape.boundaryList[-1]
            shape.freeList = np.array([])
            x =  np.concatenate((x, shape.boundaryNodes[:, 0]))
            y =  np.concatenate((y, shape.boundaryNodes[:, 1]))

        else:
            # print(shape.boundaryNodes)
            # print(shape.freeNodes)
            shape.boundaryList = np.arange(nn+1, nn+len(shape.boundaryNodes)+1, 1)
            nn = shape.boundaryList[-1]
            if len(shape.freeNodes) == 0:
                shape.freeList = np.array([])
            else:
                shape.freeList = np.arange(nn+1, nn+len(shape.freeNodes)+1, 1)
                nn = shape.freeList[-1]
            x =  np.concatenate((x, shape.boundaryNodes[:, 0], shape.freeNodes[:, 0]))
            y =  np.concatenate((y, shape.boundaryNodes[:, 1], shape.freeNodes[:, 1]))

    xy = np.transpose([x, y])
    dt = delaunay(xy)
    mesh = Mesh(x, y, dt.simplices)
    
    # Generate element list in each shape
    ne = len(dt.simplices[:,0])
    parentPolygons = []
    childPolygons = []
    for shape in shapes:
        shape.elementList = []
        if shape.isClosed:
            parentPolygons.append(ShapelyPolygon(shape.boundaryNodes))
            childPolygon = []
            for i in range(len(shape.child)):
                if shapes[shape.child[i]].shape == "point":
                    childPolygon.append([])
                else:
                    childPolygon.append(ShapelyPolygon(shapes[shape.child[i]].boundaryNodes))
            childPolygons.append(childPolygon)
        else:
            parentPolygons.append([])
            childPolygons.append([])

    cm = dt.simplices
    for i in range(ne):
        ex = x[cm[i]]
        ey = y[cm[i]]
        cc = circumCenter(ex, ey)
        point = ShapelyPoint(cc[0], cc[1])

        for j in range(len(shapes)):
            if shapes[j].isClosed:
                parentPolygon = parentPolygons[j]
                childPolygon = childPolygons[j]
                if parentPolygon.contains(point):
                    pointOutsideChild = True
                    for k in range(len(childPolygon)):
                        if shapes[shapes[j].child[k]].shape != 'point':
                            if childPolygon[k].contains(point):
                                pointOutsideChild = False
                                break

                    if pointOutsideChild:
                        shapes[j].elementList.append(i)

    # Elements
    mesh.getNeiElements()
    # Nodes
    mesh.getNeiNodes()

    return mesh

def assignSources(shapes, mesh):
    x = mesh.x
    y = mesh.y
    nn = len(x)
    # nodeSource is for both plotting and matrix assignment
    # elementSource may serve some purpose in different physics
    nodeSource = np.zeros((nn,1))
    ne = len(mesh.conmat[:,0])
    elementSource = np.zeros((ne,1))
    print("Assigning sources")
    for shape in shapes:
        if not shape.isClosed:
            nodes = shape.boundaryList
            for node in nodes:
                nodeSource[node] = shape.source/len(nodes)
                elements = mesh.neiElements[node]
                for element in elements:
                    elementSource[element] = elementSource[element] + shape.source/len(elements)
        else:
            elements = shape.elementList
            for element in elements:
                ex = []
                ey = []
                for node in mesh.conmat[element]:
                    ex.append(x[node])
                    ey.append(y[node])
                p = [ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]]
                q = [ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]]
                ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
                # if ea == 0:
                #     print("element area is zero at: ", element)
                elementSource[element] = shape.source*ea
            for element in elements:
                for node in mesh.conmat[element]:
                    nodeSource[node] = nodeSource[node] + elementSource[element]/3
    mesh.elementSource = elementSource            
    mesh.nodeSource = nodeSource
    
def assignMaterials(shapes, mesh):
    # nodeMaterial may be used for plotting
    # elementMaterial is for matrix computation
    x = mesh.x
    nn = len(x)
    nodeMaterial = np.ones((nn,1))
    ne = len(mesh.conmat[:,0])
    elementMaterial = np.ones((ne,1))
    print("Assigning material properties")
    for shape in shapes:
        if not shape.isClosed:
            nodes = shape.boundaryList
            for node in nodes:
                nodeMaterial[node] = shape.material
        else:
            nodes = np.concatenate((shape.boundaryList, shape.freeList)).astype(int)
            for node in nodes:
                nodeMaterial[node] = shape.material
        elements = shape.elementList
        for element in elements:
            elementMaterial[element] = shape.material
    mesh.nodeMaterial = nodeMaterial
    mesh.elementMaterial = elementMaterial




from matplotlib import rcParams
rcParams['font.family'] = 'Times New Roman'
import matplotlib.pyplot as plt

# import matplotlib.pyplot as plt
import matplotlib.tri as mtri
import numpy as np
from matplotlib.ticker import FormatStrFormatter

def plotGeometry(shapes, filename, dpi):
    plt.close("all")
    # fig, ax=plt.subplots(dpi=dpi)
    for shape in shapes:
        shape.show(plt)
    plt.gca().set_aspect("equal")
    plt.xlabel("x-axis [m]", fontsize= 10, fontweight='bold')
    plt.ylabel("y-axis [m]", fontsize= 11, fontweight='bold')
    plt.xticks(fontsize= 10, fontweight='bold')
    plt.yticks(fontsize= 10, fontweight='bold')
    plt.savefig(filename, dpi=dpi)

def plotGeometryNodes(shapes, filename, dpi):
    plt.close("all")
    # fig, ax=plt.subplots(dpi=dpi)
    for shape in shapes:
        shape.showNodes(plt)
    plt.gca().set_aspect("equal")
    plt.xlabel("x-axis [m]", fontsize= 10, fontweight='bold')
    plt.ylabel("y-axis [m]", fontsize= 11, fontweight='bold')
    plt.xticks(fontsize= 10, fontweight='bold')
    plt.yticks(fontsize= 10, fontweight='bold')
    plt.savefig(filename, dpi=dpi)

def plotMesh(shapes, mesh, filename, dpi):
    plt.close("all")
    # fig, ax=plt.subplots(dpi=dpi)
    # for shape in shapes:
    #     shape.show(plt)
    plt.triplot(mesh.x, mesh.y, mesh.conmat, c = 'k', lw = 100/dpi)
    plt.gca().set_aspect("equal")
    plt.xlabel("x-axis [m]", fontsize= 10, fontweight='bold')
    plt.ylabel("y-axis [m]", fontsize= 11, fontweight='bold')
    plt.xticks(fontsize= 10, fontweight='bold')
    plt.yticks(fontsize= 10, fontweight='bold')
    plt.savefig(filename, dpi=dpi)

def closeFigure():
    plt.close()

def showPartialMesh(x, y, cm, ms = 1000):
    plt.close("all")
    fig = plt.figure()
    timer = fig.canvas.new_timer(interval = ms) #creating a timer object and setting an interval of 3000 milliseconds
    timer.add_callback(closeFigure)
    plt.triplot(x, y, cm, c = 'k')
    plt.gca().set_aspect("equal")
    timer.start()
    plt.show()

def plotSources(mesh, solver, filename, dpi):
    plt.close("all")
    # fig, ax=plt.subplots(dpi=dpi)
    triang = mtri.Triangulation(mesh.x, mesh.y, mesh.conmat)
    if solver.family == "thermal":
        cmap = "hot"
    else:
        cmap = "viridis"
    plt.tricontourf(triang, mesh.nodeSource[:,0], 900, cmap = cmap)
    plt.gca().set_aspect("equal")
    plt.xlabel("x-axis [m]", fontsize= 10, fontweight='bold')
    plt.ylabel("y-axis [m]", fontsize= 11, fontweight='bold')
    plt.xticks(fontsize= 10, fontweight='bold')
    plt.yticks(fontsize= 10, fontweight='bold')
    plt.savefig(filename, dpi=dpi)

def plotMaterials(mesh, solver, filename, dpi):
    plt.close("all")
    # fig, ax=plt.subplots(dpi=dpi)
    triang = mtri.Triangulation(mesh.x, mesh.y, mesh.conmat)
    if solver.family == "thermal":
        cmap = "hot"
    else:
        cmap = "viridis"
    plt.tricontourf(triang, mesh.nodeMaterial[:,0], 900, cmap = cmap)
    plt.gca().set_aspect("equal")
    plt.xlabel("x-axis [m]", fontsize= 10, fontweight='bold')
    plt.ylabel("y-axis [m]", fontsize= 11, fontweight='bold')
    plt.xticks(fontsize= 10, fontweight='bold')
    plt.yticks(fontsize= 10, fontweight='bold')
    plt.savefig(filename, dpi=dpi)

def plotSolution(mesh, solver, solution, solutionName, withField, dpi):
    plt.close("all")
    
    if solver.family == "thermal":
        cmap = "hot"
    else:
        cmap = "viridis"
    
    if withField:
        triang = mtri.Triangulation(mesh.x, mesh.y, mesh.conmat)
        plt.tricontourf(triang, solution[:,0]-solution[:,0][0], 900, cmap = cmap)
        # plt.tricontourf(triang, np.sqrt(solver.Ex**2+solver.Ey**2), 900, cmap = cmap)
        plt.quiver(mesh.x, mesh.y, solver.Ex, solver.Ey, scale = 2)
        plt.gca().set_aspect("equal")
        plt.xlabel("x-axis [m]", fontsize= 10, fontweight='bold')
        plt.ylabel("y-axis [m]", fontsize= 11, fontweight='bold')
        plt.xticks(fontsize= 10, fontweight='bold')
        plt.yticks(fontsize= 10, fontweight='bold')
    else:
        ax = plt.axes(projection='3d')
        ax.plot_trisurf(mesh.x, mesh.y, np.transpose(solution)[0]-np.transpose(solution)[0][0], cmap = cmap, edgecolor = 'grey', linewidth=0, antialiased=False)
        ax.view_init(30, -110)
        ax.set_xlabel("x-axis [m]")
        ax.set_ylabel("y-axis [m]")
        ax.zaxis.set_label_position("top")
        if solver.name == "electrostatic":
            ax.set_zlabel("Potential [Volt]") #, format='%.0e'
        elif solver.name == "magnetostatic":
            ax.set_zlabel("Potential [Tesla meter]")
    
    filename = "Figures/5_Solution_"+solutionName
    plt.savefig(filename, dpi=dpi)


import numpy as np
from . import Geometry

class Polyline(Geometry.Geometry):
    def __init__(self, **kwargs):
        Geometry.Geometry.__init__(self, **kwargs)
        self.shape = "polyline"
        self.points = kwargs['points']
        self.isClosed = False

        # Scale the points
        v0 = (np.array(self.points)-np.array(self.points[0]))*float(self.scale)
        v0 = np.transpose(v0)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.points[0][0]], [self.points[0][1]]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array(self.points[0])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))

    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))
        
        



import numpy as np

def getMatrices(shapes, mesh, solver, solutionName):
    if solver.name == "structural":
        if (solutionName == "stress"):
            print("Generating Stress Coefficient Matrices")
            solver.getStressMatrices(mesh, shapes)
    else:
        if (solutionName == "approximate"):
            print("Generating Approximate Coefficient Matrices")
            solver.getApproximateMatrices(mesh, shapes)
        elif (solutionName == "numerical"):
            print("Generating Numerical Coefficient Matrix")
            solver.getNumericalMatrix(mesh)
            solver.getLinNumericalMatrix(mesh)
        elif (solutionName == "numericalApproxBound"):
            print("Generating Numerical Coefficient Matrix and Approximate boundary matrices")
            solver.getApproximateBoundaryMatrices(mesh, shapes)
            solver.getNumericalMatrix(mesh)
            solver.getLinNumericalMatrix(mesh)
        elif (solutionName == "numericalIterBound"):
            print("Generating Linear and Non-linear Numerical Coefficient Matrices")
            solver.getNumericalMatrix(mesh)
            solver.getLinNumericalMatrix(mesh)

def solve(shapes, mesh, solver, solutionName, solverMethod):

    if (solutionName == "approximate"):
        print("Solving for Approximate solution")
        A = solver.Aa.copy()
        b = solver.ba.copy()
        solver.bi = b*0
        if (solverMethod == "Scipy"):
            solver.solveApproximateMatricesScipy(A, b)
        elif (solverMethod == "Numpy"):
            solver.solveApproximateMatricesNumpy(A, b)
        elif (solverMethod == "Sadiku"):
            ui = b.copy()+1
            target = 0.001
            solver.solveApproximateMatricesSadiku(A, ui, b, target)
        

    elif (solutionName == "numerical"):
        print("Solving for Numerical solution")
        ep0 = solver.physicalConstant

        if solver.name == "electrostatic":
            b = mesh.nodeSource.copy()/ep0
        elif solver.name == "magnetostatic":
            b = mesh.nodeSource.copy()*ep0
        elif solver.name == "thermoostatic":
            b = mesh.nodeSource.copy()*ep0

        A = solver.An.copy()

        # Dirichelet Boundary conditions
        for i in range(len(shapes)):
            if len(shapes[i].parent) == 0:
                gs = i
                break

        lbn = shapes[gs].boundaryList
        lbn = lbn.astype(int)
        for i in lbn:
            A[i] = 0*A[i].copy()
            A[i, i] = 1
            b[i] = 0

        if (solverMethod == "Scipy"):
            solver.solveNumericalMatricesScipy(A, b)
        elif (solverMethod == "Numpy"):
            solver.solveNumericalMatricesNumpy(A, b)
        elif (solverMethod == "Sadiku"):
            ui = b.copy()+1
            target = 0.001
            solver.solveNumericalMatricesSadiku(A, ui, b, target)

        bind = np.matmul(solver.AnLin.copy(),solver.un.copy())
        # Get list of material nodes
        matNodes = np.array([])
        for shape in shapes:
            if shape.material != 1:
                matNodes = np.append([matNodes], [shape.boundaryList])
                matNodes = np.append([matNodes], [shape.freeList])

        bi = bind.copy()*0
        matNodes = matNodes.astype(int)
        for node in matNodes:
            bi[node] = bind[node].copy()

        solver.bi = bi.copy()

    elif (solutionName == "numericalApproxBound"):
        print("Solving for Numerical solution with approximate boundary values")
        Ab = solver.Aab.copy()
        bb = solver.bab.copy()
        solver.solveApproximateBoundaryMatricesScipy(Ab, bb)
        uab = solver.uab.copy()
        ep0 = solver.physicalConstant

        if solver.name == "electrostatic":
            b = mesh.nodeSource.copy()/ep0
        elif solver.name == "magnetostatic":
            b = mesh.nodeSource.copy()*ep0
        A = solver.An.copy()

        # Dirichelet Boundary conditions
        for i in range(len(shapes)):
            if len(shapes[i].parent) == 0:
                gs = i
                break

        lbn = shapes[gs].boundaryList

        for i in lbn:
            A[i] = 0*A[i].copy()
            A[i, i] = 1
            b[i] = uab[i].copy()

        if (solverMethod == "Scipy"):
            solver.solveNumericalMatricesApproxBoundScipy(A, b)
        elif (solverMethod == "Numpy"):
            solver.solveNumericalMatricesApproxBoundNumpy(A, b)
        elif (solverMethod == "Sadiku"):
            ui = b.copy()+1
            target = 0.001
            solver.solveNumericalMatricesApproxBoundSadiku(A, ui, b, target)

        bind = np.matmul(solver.AnLin.copy(),solver.unab.copy())
        # Get list of material nodes
        matNodes = np.array([])
        for shape in shapes:
            if shape.material != 1:
                matNodes = np.append([matNodes], [shape.boundaryList])
                matNodes = np.append([matNodes], [shape.freeList])

        bi = bind.copy()*0
        matNodes = matNodes.astype(int)
        for node in matNodes:
            bi[node] = bind[node].copy()

        solver.bi = bi.copy()

    elif (solutionName == "numericalIterBound"):
        print("Solving for Numerical solution with iterative boundary values")
        ep0 = solver.physicalConstant

        if solver.name == "electrostatic":
            ba = mesh.nodeSource.copy()/ep0
        elif solver.name == "magnetostatic":
            ba = mesh.nodeSource.copy()*ep0

        A = solver.An.copy()
        bi = ba.copy()*0

        # Get list of material nodes
        matNodes = np.array([])
        for shape in shapes:
            if shape.material != 1:
                matNodes = np.append([matNodes], [shape.boundaryList])
                matNodes = np.append([matNodes], [shape.freeList])

        matNodes = matNodes.astype(int)
        maxIt = 10

        for it in range(maxIt):
            bt = ba + bi
            print("Iteration of Boundary Potential", it, end='\r')
            # Boundary potential due to total sources
            solver.getIterBoundaryMatrices(mesh, shapes, bt)
            Ab = solver.Aib.copy()
            bb = solver.bib.copy()
            solver.solveIterBoundaryMatricesScipy(Ab, bb)
            uib = solver.uib.copy()

            # Dirichelet Boundary conditions
            for i in range(len(shapes)):
                if len(shapes[i].parent) == 0:
                    gs = i
                    break

            lbn = shapes[gs].boundaryList
            
            for i in lbn:
                A[i] = 0*A[i].copy()
                A[i, i] = 1
                ba[i] = uib[i].copy()

            if (solverMethod == "Scipy"):
                solver.solveNumericalMatricesApproxIterBoundScipy(A, ba)
            elif (solverMethod == "Numpy"):
                solver.solveNumericalMatricesApproxIterBoundNumpy(A, ba)
            elif (solverMethod == "Sadiku"):
                ui = b.copy()+1
                target = 0.001
                solver.solveNumericalMatricesApproxIterBoundSadiku(A, ui, ba, target)

            bind = np.matmul(solver.AnLin.copy(),solver.unib.copy())

            for node in matNodes:
                bi[node] = bind[node].copy()

        print()
        solver.bi = bi.copy()
        
    solver.findVectorFields(mesh, solver)

def getDistributedSources(shapes, mesh, solver, tn):

    tri = mesh.tri
    x = mesh.rx
    y = mesh.ry
    ep0 = solver.physicalConstant
    # pc = solver.proportionalityConstant
    if solver.name == "electrostatic":
        b = mesh.nodeSource.copy()/ep0
    elif solver.name == "magnetostatic":
        b = mesh.nodeSource.copy()*ep0
    # List of all truncated nodes
    bcopy = b.copy()
    tNodes = np.array([])

    nodeAttNames = mesh.nodeAttNames.copy()
    eleAttNames = mesh.eleAttNames.copy()
    # farnodeAttNames = mesh.farnodeAttNames.copy()

    neiNodesTn = getattr(mesh, nodeAttNames[tn])
    neiElementsTn = getattr(mesh, eleAttNames[tn])
    # farneiNodesTn = getattr(mesh, farnodeAttNames[tn-1])

    for domain in shapes:
        if domain.source != 0:

            if domain.shape == "point":
                domainNode = domain.boundaryList[0]
                tNodes = np.append(tNodes, neiNodesTn[domainNode])

            else:
                domainNodes = domain.boundaryList
                domainNodes = np.append(domainNodes, domain.freeList)
                domainNodes = domainNodes.astype(int)
                for node in domainNodes:
                    tNodes = np.append(tNodes, neiNodesTn[node])

    tNodes = np.unique(tNodes)
    tNodes = tNodes.astype(int)

    # count = 0
    for node in tNodes:

        miniNodes = neiNodesTn[node]
        miniElements = neiElementsTn[node]
        miniTri = tri[miniElements]

        miniNodes = np.sort(miniNodes)

        for j in range(len(miniNodes)):
            if miniNodes[j] == node:
                miniMeshCentreNode = j
                break

        for j in range(len(miniElements)):
            for k in range(3):
                globalNode = miniTri[j,k]

                for l in range(len(miniNodes)):
                    if globalNode == miniNodes[l]:
                        miniTri[j,k] = l
                        break

        minix = x[miniNodes].copy()
        miniy = y[miniNodes].copy()

        mne = len(miniElements)
        mnn = len(minix)

        miniA = np.zeros([mnn, mnn])
        minib = np.zeros([mnn,1])

        for i in range(mne):

            # 0-1 edge
            k = 0
            l = 1
            Ex = 0
            Ey = 0

            for mnode in range(len(miniNodes)):
                px = minix[mnode].copy()
                py = miniy[mnode].copy()
                Rx = (minix[miniTri[i,k]] + minix[miniTri[i,l]])/2 - px
                Ry = (miniy[miniTri[i,k]] + miniy[miniTri[i,l]])/2 - py
                Ex = Ex + b[miniNodes[mnode]]/(2*np.pi)*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                Ey = Ey + b[miniNodes[mnode]]/(2*np.pi)*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = minix[miniTri[i,l]] - minix[miniTri[i,k]]
            ry = miniy[miniTri[i,l]] - miniy[miniTri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            miniA[miniTri[i,k], miniTri[i,k]] = miniA[miniTri[i,k], miniTri[i,k]] +1
            miniA[miniTri[i,k], miniTri[i,l]] = -1
            minib[miniTri[i,k]]           = minib[miniTri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 1-2 edge
            k = 1
            l = 2
            Ex = 0
            Ey = 0

            for mnode in range(len(miniNodes)):
                px = minix[mnode].copy()
                py = miniy[mnode].copy()
                Rx = (minix[miniTri[i,k]] + minix[miniTri[i,l]])/2 - px
                Ry = (miniy[miniTri[i,k]] + miniy[miniTri[i,l]])/2 - py
                Ex = Ex + b[miniNodes[mnode]]/(2*np.pi)*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                Ey = Ey + b[miniNodes[mnode]]/(2*np.pi)*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = minix[miniTri[i,l]] - minix[miniTri[i,k]]
            ry = miniy[miniTri[i,l]] - miniy[miniTri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            miniA[miniTri[i,k], miniTri[i,k]] = miniA[miniTri[i,k], miniTri[i,k]] +1
            miniA[miniTri[i,k], miniTri[i,l]] = -1
            minib[miniTri[i,k]]           = minib[miniTri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 2-0 edge
            k = 2
            l = 0
            Ex = 0
            Ey = 0

            for mnode in range(len(miniNodes)):
                px = minix[mnode].copy()
                py = miniy[mnode].copy()
                Rx = (minix[miniTri[i,k]] + minix[miniTri[i,l]])/2 - px
                Ry = (miniy[miniTri[i,k]] + miniy[miniTri[i,l]])/2 - py
                Ex = Ex + b[miniNodes[mnode]]/(2*np.pi)*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                Ey = Ey + b[miniNodes[mnode]]/(2*np.pi)*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = minix[miniTri[i,l]] - minix[miniTri[i,k]]
            ry = miniy[miniTri[i,l]] - miniy[miniTri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            miniA[miniTri[i,k], miniTri[i,k]] = miniA[miniTri[i,k], miniTri[i,k]] +1
            miniA[miniTri[i,k], miniTri[i,l]] = -1
            minib[miniTri[i,k]]           = minib[miniTri[i,k]] + Er*np.sqrt(rx**2+ry**2)

        solver.solveminiNumpy(miniA, minib)

        miniu = solver.miniu
        solver.getMiniLinNumericalMatrix(minix, miniy, miniTri)
        miniAn = solver.miniAn

        minibnshould = np.matmul(miniAn, miniu)
        print(node, b[node], minibnshould[miniMeshCentreNode])
        bcopy[node] = minibnshould[miniMeshCentreNode]

    solver.bn = b.copy()
    solver.bdss = bcopy.copy()
    mesh.nodeSourceDss = bcopy.copy()*ep0

    print()

def solveDSS(shapes, mesh, solver, solutionName, solverMethod):

    if (solutionName == "approximate"):
        print("DSS does not support approximate solution")

    elif (solutionName == "numerical"):
        print("Solving for Numerical solution")
        #ep0 = solver.physicalConstant

        b = solver.bdss.copy()
        A = solver.An.copy()

        # Dirichelet Boundary conditions
        lbn = shapes[0].boundaryList

        for i in lbn:
            A[i] = 0*A[i].copy()
            A[i, i] = 1
            b[i] = 1

        if (solverMethod == "Scipy"):
            solver.solveNumericalMatricesScipy(A, b)
        elif (solverMethod == "Numpy"):
            solver.solveNumericalMatricesNumpy(A, b)
        elif (solverMethod == "Sadiku"):
            ui = b.copy()+1
            target = 0.001
            solver.solveNumericalMatricesSadiku(A, ui, b, target)

    elif (solutionName == "numericalApproxBound"):
        print("Solving for Numerical solution with approximate boundary values")
        Ab = solver.Aab.copy()
        bb = solver.bab.copy()
        solver.solveApproximateBoundaryMatricesScipy(Ab, bb)
        uab = solver.uab.copy()

        # ep0 = solver.physicalConstant

        b = solver.bdss
        A = solver.An.copy()

        # Dirichelet Boundary conditions
        lbn = shapes[0].boundaryList

        for i in lbn:
            A[i] = 0*A[i].copy()
            A[i, i] = 1
            b[i] = uab[i].copy()

        if (solverMethod == "Scipy"):
            solver.solveNumericalMatricesApproxBoundScipy(A, b)
        elif (solverMethod == "Numpy"):
            solver.solveNumericalMatricesApproxBoundNumpy(A, b)
        elif (solverMethod == "Sadiku"):
            ui = b.copy()+1
            target = 0.001
            solver.solveNumericalMatricesApproxBoundSadiku(A, ui, b, target)

    elif (solutionName == "numericalIterBound"):
        print("Solving for Numerical solution with iterative boundary values")
        ep0 = solver.physicalConstant

        if solver.name == "electrostatic":
            ba = mesh.nodeSource.copy()/ep0
        elif solver.name == "magnetostatic":
            ba = mesh.nodeSource.copy()*ep0

        bdss = solver.bdss.copy()

        A = solver.An.copy()
        bi = ba.copy()*0

        # Get list of material nodes
        matNodes = np.array([])
        for domain in shapes:
            if domain.material != 1:
                matNodes = np.append([matNodes], [domain.boundaryList])
                matNodes = np.append([matNodes], [domain.freeList])

        matNodes = matNodes.astype(int)

        maxIt = 10

        for it in range(maxIt):
            bt = ba + bi
            print("Iteration of Boundary Potential", it, end='\r')

            # Boundary potential due to total sources
            solver.getIterBoundaryMatrices(mesh, shapes, bt)
            Ab = solver.Aib.copy()
            bb = solver.bib.copy()
            solver.solveIterBoundaryMatricesScipy(Ab, bb)
            uib = solver.uib.copy()

            # Dirichelet Boundary conditions
            lbn = shapes[0].boundaryList

            for i in lbn:
                A[i] = 0*A[i].copy()
                A[i, i] = 1
                bdss[i] = uib[i].copy()

            if (solverMethod == "Scipy"):
                solver.solveNumericalMatricesApproxIterBoundScipy(A, bdss)
            elif (solverMethod == "Numpy"):
                solver.solveNumericalMatricesApproxIterBoundNumpy(A, bdss)
            elif (solverMethod == "Sadiku"):
                ui = b.copy()+1
                target = 0.001
                solver.solveNumericalMatricesApproxIterBoundSadiku(A, ui, bdss, target)

            bind = np.matmul(solver.AnLin.copy(),solver.unib.copy())

            for node in matNodes:
                bi[node] = bind[node].copy()
            
        print()
        solver.bi = bi.copy()

    solver.findVectorFields(mesh, solver)



class Geometry:

    def __init__(self, **kwargs):
        self.stroke = kwargs.get('stroke', "black")
        self.strokeWidth = kwargs.get('strokeWidth', "1")
        self.fill = kwargs.get('fill', "transparent")
        self.non = kwargs.get('non', "0")
        self.relativeDensity = kwargs.get('relativeDensity', "[1, 1]")
        self.scale = kwargs.get('scale', "1")
        self.rotate = kwargs.get('rotate', "0")
        self.pivot = kwargs.get('pivot', "itself")
        self.pivotAngle = kwargs.get('pivotAngle', "0")
        self.parent = []
        self.child = []
        self.isParent = True
        self.isChild = True
        self.isHole = kwargs.get('isHole', False) # Work on it later
        self.source = 0
        self.material = 1
    


import numpy as np
from shapely.geometry.polygon import Polygon as ShapelyPolygon
from shapely.geometry import Point as ShapelyPoint
from .Polygon import Polygon

def add(shape1, shape2):

    if shape1.isClosed and shape2.isClosed: # Add only if both are closed
        points1, points2 = shape1.boundaryNodes, shape2.boundaryNodes
        points1io, points2io = [], []
        
        polygon = ShapelyPolygon(points2) # separating interior and exterior points of shape 1
        for i in range(len(points1)):
            point = ShapelyPoint(points1[i])
            if polygon.contains(point):
                points1io.append(1)
            else:
                points1io.append(0)
        npo1 = len(points1) - sum(points1io)

        polygon = ShapelyPolygon(points1) # separating interior and exterior points of shape 2
        for i in range(len(points2)):
            point = ShapelyPoint(points2[i])
            if polygon.contains(point):
                points2io.append(1)
            else:
                points2io.append(0)
        npo2 = len(points2) - sum(points2io)

        if (npo1 == len(points1) and npo2 == len(points2)) or (npo1 == 0 and npo2 == len(points2)) or (npo1 == len(points1) and npo2 == 0):
            raise ValueError("Cannot add non overlapping or completely inclusive shapes.")
        else:
            segmentLengths = []
            for i in range(len(points1)-1):
                segmentLengths.append(np.sqrt((points1[i][0]-points1[i+1][0])**2+(points1[i][1]-points1[i+1][1])**2))
            for i in range(len(points2)-1):
                segmentLengths.append(np.sqrt((points2[i][0]-points2[i+1][0])**2+(points2[i][1]-points2[i+1][1])**2))

            # Rotate index
            for i in range(len(points1)):
                if points1io[i] == 1 and points1io[i+1] == 0:
                    rotationIndex = i+1
                    break
            points1 = np.roll(points1, (-rotationIndex,-rotationIndex))

            for i in range(len(points2)):
                if points2io[i] == 1 and points2io[i+1] == 0:
                    rotationIndex = i+1
                    break
            points2 = np.roll(points2, (-rotationIndex,-rotationIndex))

            # vstack outside points
            boundaryNodes = np.vstack([points1[0:npo1], points2[0:npo2]])

            # remove short segments from the union
            minSeglength = min(segmentLengths)
            pointsToRemove = []
            i = npo1-1
            if np.sqrt((boundaryNodes[i][0]-boundaryNodes[i+1][0])**2+(boundaryNodes[i][1]-boundaryNodes[i+1][1])**2) < 0.5*minSeglength:
                pointsToRemove.append(i)
            i = npo1+npo2-2
            if np.sqrt((boundaryNodes[0][0]-boundaryNodes[i+1][0])**2+(boundaryNodes[0][1]-boundaryNodes[i+1][1])**2) < 0.5*minSeglength:
                pointsToRemove.append(i+1)

            boundaryNodes = np.delete(boundaryNodes, pointsToRemove, axis=0)
            return Polygon(points = boundaryNodes, stroke = shape1.stroke)
    else:
        raise ValueError("Cannot add open shapes.")

def intersect(shape1, shape2):

    if shape1.isClosed and shape2.isClosed: # Intersect only if both are closed
        points1, points2 = shape1.boundaryNodes, shape2.boundaryNodes
        points1io, points2io = [], []
        
        polygon = ShapelyPolygon(points2) # separating interior and exterior points of shape 1
        for i in range(len(points1)):
            point = ShapelyPoint(points1[i])
            if polygon.contains(point):
                points1io.append(1)
            else:
                points1io.append(0)
        npi1 = sum(points1io)
        npo1 = len(points1) - sum(points1io)

        polygon = ShapelyPolygon(points1) # separating interior and exterior points of shape 2
        for i in range(len(points2)):
            point = ShapelyPoint(points2[i])
            if polygon.contains(point):
                points2io.append(1)
            else:
                points2io.append(0)
        npi2 = sum(points2io)
        npo2 = len(points2) - sum(points2io)

        if (npo1 == len(points1) and npo2 == len(points2)) or (npo1 == 0 and npo2 == len(points2)) or (npo1 == len(points1) and npo2 == 0):
            raise ValueError("Cannot intersect non overlapping or completely inclusive shapes.")
        else:
            segmentLengths = []
            for i in range(len(points1)-1):
                segmentLengths.append(np.sqrt((points1[i][0]-points1[i+1][0])**2+(points1[i][1]-points1[i+1][1])**2))
            for i in range(len(points2)-1):
                segmentLengths.append(np.sqrt((points2[i][0]-points2[i+1][0])**2+(points2[i][1]-points2[i+1][1])**2))

            # Rotate index
            for i in range(len(points1)):
                if points1io[i] == 0 and points1io[i+1] == 1:
                    rotationIndex = i+1
                    break
            points1 = np.roll(points1, (-rotationIndex,-rotationIndex))

            for i in range(len(points2)):
                if points2io[i] == 0 and points2io[i+1] == 1:
                    rotationIndex = i+1
                    break
            points2 = np.roll(points2, (-rotationIndex,-rotationIndex))

            # vstack inside points
            boundaryNodes = np.vstack([points1[0:npi1], points2[0:npi2]])

            # remove short segments from the intersection
            minSeglength = min(segmentLengths)
            pointsToRemove = []
            i = npi1-1
            if np.sqrt((boundaryNodes[i][0]-boundaryNodes[i+1][0])**2+(boundaryNodes[i][1]-boundaryNodes[i+1][1])**2) < 0.5*minSeglength:
                pointsToRemove.append(i)
            i = npi1+npi2-2
            if np.sqrt((boundaryNodes[0][0]-boundaryNodes[i+1][0])**2+(boundaryNodes[0][1]-boundaryNodes[i+1][1])**2) < 0.5*minSeglength:
                pointsToRemove.append(i+1)

            boundaryNodes = np.delete(boundaryNodes, pointsToRemove, axis=0)
            return Polygon(points = boundaryNodes, stroke = shape1.stroke)
    else:
        raise ValueError("Cannot intersect open shapes.")

def subtract(shape1, shape2):

    if shape1.isClosed and shape2.isClosed: # Subtract only if both are closed
        points1, points2 = shape1.boundaryNodes, shape2.boundaryNodes
        points1io, points2io = [], []
        
        polygon = ShapelyPolygon(points2) # separating interior and exterior points of shape 1
        for i in range(len(points1)):
            point = ShapelyPoint(points1[i])
            if polygon.contains(point):
                points1io.append(1)
            else:
                points1io.append(0)
        npi1 = sum(points1io)
        npo1 = len(points1) - sum(points1io)

        polygon = ShapelyPolygon(points1) # separating interior and exterior points of shape 2
        for i in range(len(points2)):
            point = ShapelyPoint(points2[i])
            if polygon.contains(point):
                points2io.append(1)
            else:
                points2io.append(0)
        npo2 = len(points2) - sum(points2io)

        if (npo1 == len(points1) and npo2 == len(points2)) or (npo1 == 0 and npo2 == len(points2)) or (npo1 == len(points1) and npo2 == 0):
            raise ValueError("Cannot subtract non overlapping or completely inclusive shapes.")
        else:
            segmentLengths = []
            for i in range(len(points1)-1):
                segmentLengths.append(np.sqrt((points1[i][0]-points1[i+1][0])**2+(points1[i][1]-points1[i+1][1])**2))
            for i in range(len(points2)-1):
                segmentLengths.append(np.sqrt((points2[i][0]-points2[i+1][0])**2+(points2[i][1]-points2[i+1][1])**2))

            # Rotate
            for i in range(len(points1)):
                if points1io[i] == 0 and points1io[i+1] == 1:
                    rotationIndex = i+1
                    break
            points1 = np.roll(points1, (-rotationIndex,-rotationIndex))

            for i in range(len(points2)):
                if points2io[i] == 1 and points2io[i+1] == 0:
                    rotationIndex = i+1
                    break
            points2 = np.roll(points2, (-rotationIndex,-rotationIndex))

            # vstack flipped inside points with unflipped outside points 
            boundaryNodes = np.vstack([np.flip(points1[0:npi1], 0), points2[0:npo2]])

            # remove short segments from the subtraction
            minSeglength = min(segmentLengths)
            pointsToRemove = []
            i = npi1-1
            if np.sqrt((boundaryNodes[i][0]-boundaryNodes[i+1][0])**2+(boundaryNodes[i][1]-boundaryNodes[i+1][1])**2) < 0.5*minSeglength:
                pointsToRemove.append(i)
            i = npi1+npo2-2
            if np.sqrt((boundaryNodes[0][0]-boundaryNodes[i+1][0])**2+(boundaryNodes[0][1]-boundaryNodes[i+1][1])**2) < 0.5*minSeglength:
                pointsToRemove.append(i+1)

            boundaryNodes = np.delete(boundaryNodes, pointsToRemove, axis=0)
            return Polygon(points = boundaryNodes, stroke = shape2.stroke)
    else:
        raise ValueError("Cannot subtract open shapes.")

# def split(shape1, shape2): # Provide provision to split with multiple shapes and domain tags for each segment.




from .Solver import Solver
import numpy as np

class Structural(Solver):
    def __init__(self):
        Solver.__init__(self)
        self.name = "structural"
        self.physicalConstant = 0
        self.proportionalityConstant = 0

    def getStressMatrices(self, mesh, initialDomains):

        x = mesh.rx
        y = mesh.ry
        tri = mesh.tri


# Import packages
import joblib
from . import PlotUtils
from . import Mesh
from .Solvers import Solve

def mesh(crse = 1, dpi = 300, meshingMethod = "ruppert"):
    print("Generating mesh")
    geometry = joblib.load("Data/geometry.sav")
    Mesh.prepareGeometryForMeshing(geometry)
    Mesh.generateNodes(geometry, crse, meshingMethod)
    mesh = Mesh.createMesh(geometry)
    joblib.dump(geometry, "Data/geometry.sav")
    joblib.dump(mesh, "Data/mesh.sav")
    if dpi < 300:
        dpi = 300
        print("Value of dpi too low. Resetting dpi to 300.")
    PlotUtils.plotMesh(geometry, mesh, "Figures/2_Mesh.png", dpi)

def assigedSources(dpi):
    shapes = joblib.load("Data/geometry.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Mesh.assignSources(shapes, mesh)
    joblib.dump(mesh, "Data/mesh.sav")
    PlotUtils.plotSources(mesh, solver, "Figures/3_Sources.png", dpi)

def assigedMaterials(dpi):
    shapes = joblib.load("Data/geometry.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Mesh.assignMaterials(shapes, mesh)
    joblib.dump(mesh, "Data/mesh.sav")
    PlotUtils.plotMaterials(mesh, solver, "Figures/4_Materials.png", dpi)
    
def solution(solutionName = "numerical", solverMethod = "Scipy", withDifference = False, withField = False, dpi = 300):
    shapes = joblib.load("Data/geometry.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Solve.getMatrices(shapes, mesh, solver, solutionName)
    Solve.solve(shapes, mesh, solver, solutionName, solverMethod)
    if dpi < 300:
        dpi = 300
        print("Value of dpi too low. Resetting dpi to 300.")
    showSolution(mesh, solver, solutionName, withField, dpi)
    if withDifference:
        showDifference(mesh, solver, solutionName, dpi)
    joblib.dump(mesh, "Data/mesh.sav")
    joblib.dump(solver, "Data/solver.sav")

def showSolution(mesh, solver, solutionName, withField, dpi):
    if (solutionName == "approximate"):
        solution = solver.ua
        PlotUtils.plotSolution(mesh, solver, solution, solutionName, withField, dpi)
    elif (solutionName == "numerical"):
        solution = solver.un
        PlotUtils.plotSolution(mesh, solver, solution, solutionName, withField, dpi)
    elif (solutionName == "numericalApproxBound"):
        solution = solver.unab
        PlotUtils.plotSolution(mesh, solver, solution, solutionName, withField, dpi)
    elif (solutionName == "numericalIterBound"):
        solution = solver.unib
        PlotUtils.plotSolution(mesh, solver, solution, solutionName, withField, dpi)

def showDifference(mesh, solver, solutionName, dpi):
    print()


import numpy as np
from . import Geometry

class Line(Geometry.Geometry):
    def __init__(self, **kwargs):
        Geometry.Geometry.__init__(self, **kwargs)
        self.shape = "line"
        self.x1 = kwargs['x1']
        self.x2 = kwargs['x2']
        self.y1 = kwargs['y1']
        self.y2 = kwargs['y2']
        self.non = kwargs['non']
        self.isClosed = False

        non = self.non
        if self.x1 == self.x2:
            x = np.arange(int(non))*0 + self.x1
            y = np.arange(self.y1, self.y2, (self.y2-self.y1)/(non-1))
            y = np.hstack([y, self.y2])
        elif self.y1 == self.y2:
            x = np.arange(self.x1, self.x2, (self.x2-self.x1)/(non-1))
            x = np.hstack([x, self.x2])
            y = np.arange(int(non))*0 + self.y1
        else:
            x = np.arange(self.x1, self.x2, (self.x2-self.x1)/(non-1))
            x = np.hstack([x, self.x2])
            y = np.arange(self.y1, self.y2, (self.y2-self.y1)/(non-1))
            y = np.hstack([y, self.y2])

        x = x-x[0]
        y = y-y[0]
        # list of points
        v0 = np.vstack([x, y])*float(self.scale)
        # Rotate the points around itself first
        thr = float(self.rotate)
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Add Pivot if axis of rotation is not around itself
        if self.pivot != "itself":
            vr = vr + np.array([[self.x1], [self.y1]]) - np.array([[self.pivot[0]], [self.pivot[1]]])
            # Rotate around the pivot
            thr = float(self.pivotAngle)
            R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
            [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
            vr = np.dot(R,vr)
        # Reposition the points
        if self.pivot == "itself":
            position = np.array([self.x1, self.y1])
        else:
            position = np.array(self.pivot)
        boundaryNodes = position +  np.transpose(vr)
        self.boundaryNodes = boundaryNodes
        segments = []
        for i in range(len(boundaryNodes)-1):
            segments.append([i, i+1])
        self.segments = segments

    def show(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], color = self.stroke, lw = float(self.strokeWidth))
        
    def showNodes(self, plt):
        v = self.boundaryNodes
        for i in range(len(v)-1):
            plt.plot(v[[i,i+1],0], v[[i,i+1],1], ".", color = self.stroke, markersize = float(self.strokeWidth))